from __future__ import annotations

from typing import Any, ClassVar, overload

class ConstructionMethod:
    __members__: ClassVar[dict[ConstructionMethod, int]] = ...  # read-only
    recursive: ClassVar[ConstructionMethod] = ...
    sequential: ClassVar[ConstructionMethod] = ...
    @overload
    def __init__(self, value: int) -> None: ...
    @overload
    def __init__(self, arg0: str) -> None: ...
    def __eq__(self, other: object) -> bool: ...
    def __getstate__(self) -> int: ...
    def __hash__(self) -> int: ...
    def __index__(self) -> int: ...
    def __int__(self) -> int: ...
    def __ne__(self, other: object) -> bool: ...
    def __setstate__(self, state: int) -> None: ...
    @property
    def name(self) -> str: ...
    @property
    def value(self) -> int: ...

def construct(
    circ: object, method: ConstructionMethod = ..., store_dd: bool = ..., store_matrix: bool = ...
) -> dict[str, Any]: ...
def construct_grover(
    nqubits: int = ...,
    seed: int = ...,
    method: ConstructionMethod = ...,
    store_dd: bool = ...,
    store_matrix: bool = ...,
) -> dict[str, Any]: ...
def construct_qft(
    nqubits: int = ..., method: ConstructionMethod = ..., store_dd: bool = ..., store_matrix: bool = ...
) -> dict[str, Any]: ...
def matrix_from_dd(serialized_dd: str) -> dict[str, Any]: ...



#
# This file is part of MQT QFR library which is released under the MIT license.
# See file README.md or go to https://www.cda.cit.tum.de/research/quantum/ for more information.
#

from mqt.qfr.pyqfr import (
    ConstructionMethod,
    construct,
    construct_grover,
    construct_qft,
    matrix_from_dd,
)

__all__ = [
    "construct",
    "construct_grover",
    "construct_qft",
    "ConstructionMethod",
    "matrix_from_dd",
]



import numpy as np
from mqt import qecc

code = qecc.Code("/path/to/Hx", "path/to/Hz")
decoder = qecc.UFHeuristic()
decoder.set_code(code)
x_err = qecc.sample_iid_pauli_err(code.N, 0.05)
decoder.decode(code.get_x_syndrome(x_err))
result = decoder.result
print(result)
residual_err = np.array(x_err) ^ np.array(result.estimate)
print(code.is_x_stabilizer(residual_err))



from __future__ import annotations

import nox
from nox.sessions import Session

nox.options.sessions = ["lint"]


@nox.session
def lint(session: Session) -> None:
    """
    Lint the Python part of the codebase using pre-commit.
    Simply execute `nox -rs lint` to run all configured hooks.
    """
    session.install("pre-commit")
    session.run("pre-commit", "run", "--all-files", *session.posargs)


@nox.session
def pylint(session: Session) -> None:
    """
    Run pylint.
    Simply execute `nox -rs pylint` to run pylint.
    Run as `nox -rs pylint -- skip-install` to skip installing the package and its dependencies.
    """
    session.install("pylint")
    run_install = True
    if session.posargs and "skip-install" in session.posargs:
        run_install = False
        session.posargs.remove("skip-install")
    if run_install:
        session.install("-e", ".")
    session.run("pylint", "mqt.qfr", "--extension-pkg-allow-list=mqt.qfr.pyqfr", *session.posargs)


@nox.session
def mypy(session: Session) -> None:
    """
    Run mypy.
    Simply execute `nox -rs mypy` to run mypy.
    """
    session.install("pre-commit")
    session.run("pre-commit", "run", "--all-files", "--hook-stage", "manual", "mypy", *session.posargs)



# using tools from https://github.com/quantumgizmos/bp_osd
import numpy as np
from bposd.hgp import hgp
from ldpc.codes import ring_code

# toric code
for i in range(2, 100):
    h1 = ring_code(i)
    h2 = ring_code(i)

    qcode = hgp(h1, h2, compute_distance=True)
    if qcode.test:  # to make sure it is a valid css code
        np.savetxt(f"./toric_{qcode.code_params}_hz.txt", qcode.hz, fmt="%d", newline="\n")



# using tools from https://github.com/quantumgizmos/bp_osd
import ldpc.protograph as pt
import numpy as np
from lifted_hgp import lifted_hgp

a1 = pt.array([[(0), (11), (7), (12)], [(1), (8), (1), (8)], [(11), (0), (4), (8)], [(6), (2), (4), (12)]])

qcode = lifted_hgp(lift_parameter=32, a=a1, b=a1)
if qcode.test:
    np.savetxt(f"./lp_{qcode.code_params}_hz.txt", qcode.hz, fmt="%d", newline="\n")
    np.savetxt(f"./lp_{qcode.code_params}_hx.txt", qcode.hx, fmt="%d", newline="\n")



import gdb
import re

ns_pattern = re.compile(r'nlohmann(::json_abi(?P<tags>\w*)(_v(?P<v_major>\d+)_(?P<v_minor>\d+)_(?P<v_patch>\d+))?)?::(?P<name>.+)')
class JsonValuePrinter:
    "Print a json-value"

    def __init__(self, val):
        self.val = val

    def to_string(self):
        if self.val.type.strip_typedefs().code == gdb.TYPE_CODE_FLT:
            return ("%.6f" % float(self.val)).rstrip("0")
        return self.val

def json_lookup_function(val):
    m = ns_pattern.fullmatch(val.type.strip_typedefs().name)
    name = m.group('name')
    if name and name.startswith('basic_json<') and name.endswith('>'):
        m = ns_pattern.fullmatch(str(val['m_type']))
        t = m.group('name')
        if t and t.startswith('detail::value_t::'):
            try:
                union_val = val['m_value'][t.removeprefix('detail::value_t::')]
                if union_val.type.code == gdb.TYPE_CODE_PTR:
                    return gdb.default_visualizer(union_val.dereference())
                else:
                    return JsonValuePrinter(union_val)
            except Exception:
                return JsonValuePrinter(val['m_type'])

gdb.pretty_printers.append(json_lookup_function)



import ldpc.protograph as pt
import numpy as np
from bposd.hgp import hgp

# med sized HGP code from https://github.com/quantumgizmos/bp_osd

h = np.loadtxt("examples/mkmn_24_6_10.txt").astype(int)
seed_code = np.loadtxt("examples/mkmn_24_6_10.txt").astype(int)
# print(seed_code)
qcode = hgp(seed_code, compute_distance=True)
qcode.canonical_logicals()
qcode.test()
print(qcode.code_params)
print(qcode.hx)
print("hx:")
print(qcode.hx)
np.savetxt(f"./hgp_{qcode.code_params}_hx.txt", qcode.hx, fmt="%d", newline="\n")

# larger code
a1 = pt.array([[(0), (11), (7), (12)], [(1), (8), (1), (8)], [(11), (0), (4), (8)], [(6), (2), (4), (12)]])
H = a1.to_binary(lift_parameter=10)
qcode = hgp(H, H, compute_distance=True)
qcode.test()
np.savetxt(f"./hgp_{qcode.code_params}_hx.txt", qcode.hx, fmt="%d", newline="\n")



#!/usr/bin/env python3

import argparse
import itertools
import jinja2
import os
import re
import sys

def semver(v):
    if not re.fullmatch(r'\d+\.\d+\.\d+', v):
        raise ValueError
    return v

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--version', required=True, type=semver, help='Library version number')
    parser.add_argument('output', help='Output directory for nlohmann_json.natvis')
    args = parser.parse_args()

    namespaces = ['nlohmann']
    abi_prefix = 'json_abi'
    abi_tags = ['_diag', '_ldvcmp']
    version = '_v' + args.version.replace('.', '_')
    inline_namespaces = []

    # generate all combinations of inline namespace names
    for n in range(0, len(abi_tags) + 1):
        for tags in itertools.combinations(abi_tags, n):
            ns = abi_prefix + ''.join(tags)
            inline_namespaces += [ns, ns + version]

    namespaces += [f'{namespaces[0]}::{ns}' for ns in inline_namespaces]

    env = jinja2.Environment(loader=jinja2.FileSystemLoader(searchpath=sys.path[0]), autoescape=True, trim_blocks=True,
                                                            lstrip_blocks=True, keep_trailing_newline=True)
    template = env.get_template('nlohmann_json.natvis.j2')
    natvis = template.render(namespaces=namespaces)

    with open(os.path.join(args.output, 'nlohmann_json.natvis'), 'w') as f:
        f.write(natvis)



from __future__ import annotations

import os

import nox
from nox.sessions import Session

nox.options.sessions = ["lint"]

PYTHON_ALL_VERSIONS = ["3.7", "3.8", "3.9", "3.10", "3.11"]

if os.environ.get("CI", None):
    nox.options.error_on_missing_interpreters = True


@nox.session
def lint(session: Session) -> None:
    """
    Lint the Python part of the codebase using pre-commit.
    Simply execute `nox -rs lint` to run all configured hooks.
    """
    session.install("pre-commit")
    session.run("pre-commit", "run", "--all-files", *session.posargs)



#
# This file is part of MQT QECC library which is released under the MIT license.
# See file README.md for more information.
#

from mqt.qecc.pyqecc import (
    Code,
    Decoder,
    DecodingResult,
    DecodingResultStatus,
    DecodingRunInformation,
    GrowthVariant,
    UFDecoder,
    UFHeuristic,
    apply_ecc,
    sample_iid_pauli_err,
)

__all__ = [
    "Code",
    "Decoder",
    "UFHeuristic",
    "UFDecoder",
    "GrowthVariant",
    "DecodingResult",
    "DecodingResultStatus",
    "DecodingRunInformation",
    "sample_iid_pauli_err",
    "apply_ecc",
]



from __future__ import annotations

import argparse

from mqt import qecc
from qiskit import Aer, QuantumCircuit, execute, providers
from qiskit.result import counts
from qiskit_aer.noise import (
    NoiseModel,
    QuantumError,
    amplitude_damping_error,
    depolarizing_error,
)
from qiskit_aer.noise.errors import pauli_error


def compose_error(error: QuantumError, new_error: QuantumError) -> QuantumError:
    if error is None:
        error = new_error
    else:
        error = error.compose(new_error)
    return error


def create_noise_model(n_model: str, p_error: float) -> NoiseModel:
    # Create an empty noise model
    noise_model = NoiseModel()
    error = None
    for char in n_model:
        if char == "B":
            # Add a bit flip error channel
            new_error = pauli_error([("X", p_error), ("I", 1 - p_error)])
            error = compose_error(error, new_error)
        elif char == "D":
            # Add a depolarization error channel
            new_error = depolarizing_error(p_error, 1)
            error = compose_error(error, new_error)
        elif char == "A":
            # simple amplitude damping error channel which mimics energy loss to the environment (T1-error)
            new_error = amplitude_damping_error(p_error)
            error = compose_error(error, new_error)
        elif char == "P":
            # Add a phase flip error channel
            new_error = pauli_error([("Z", p_error), ("I", 1 - p_error)])
            error = compose_error(error, new_error)

        else:
            raise ValueError("Unknown error type in noise model: " + char)

    assert error is not None

    # single qubit and multi qubit operations are treated the same at the moment
    noise_model.add_all_qubit_quantum_error(
        error, ["u1", "u2", "u3", "h", "id", "t", "tdg", "sdg", "rx", "ry", "rz", "s"]
    )
    noise_model.add_all_qubit_quantum_error(error.tensor(error), ["cx", "swap"])
    noise_model.add_all_qubit_quantum_error(error.tensor(error).tensor(error), ["cswap"])

    return noise_model


def print_simulation_results(result_counts: counts, n_shots: int, threshold_probability: float = 0) -> None:
    printed_results = 0
    summarized_counts: dict[str, int] = {}
    for result_id in result_counts:
        sub_result = result_id.split(" ")[-1]
        if sub_result not in summarized_counts.keys():
            summarized_counts[sub_result] = 0
        summarized_counts[sub_result] += result_counts[result_id]

    for result_id in sorted(summarized_counts.keys()):
        # Print all results > threshold_probability
        if summarized_counts[result_id] / n_shots > threshold_probability or printed_results == 0:
            result_string = str(result_id)
            print("State |" + result_string + "> probability " + str(summarized_counts[result_id] / n_shots))
            printed_results += 1
            if printed_results == 1000:
                break


def main() -> None:
    parser = argparse.ArgumentParser(description="Qiskit wrapper for the ECC Framework")
    parser.add_argument(
        "-m",
        type=str,
        default="D",
        help="Define the error_channels (e.g., -m APD), available errors channels are amplitude "
        'damping (A), phase flip (P), bit flip (B), and depolarization (D) (Default="D")',
    )
    parser.add_argument("-p", type=float, default=0.001, help="Set the noise probability (Default=0.001)")
    parser.add_argument("-n", type=int, default=2000, help="Set the number of shots for the simulation (Default=2000)")
    parser.add_argument("-s", type=int, default=0, help="Set a seed (Default=0)")
    parser.add_argument("-f", type=str, required=True, help="Path to a OpenQASM file")
    parser.add_argument(
        "-e",
        type=str,
        required=False,
        default=None,
        help="Export circuit with applied ECC as OpenQASM circuit instead of "
        'simulating it (e.g., -e "/path/to/new/openqasm_file") (Default=None)',
    )
    parser.add_argument(
        "-fs",
        type=str,
        default="aer_simulator_stabilizer",
        help='Specify a simulator (Default="aer_simulator_stabilizer", which is fast but does not support '
        "non-Clifford gates. Available: " + str(Aer.backends()),
    )
    parser.add_argument(
        "-ecc",
        type=str,
        default="Q7Steane",
        help='Specify an ECC to be applied to the circuit. Currently available are "none", "Q3Shor", "Q5Laflamme", '
        '"Q7Steane", "Q9Shor", "Q9Surface", and "Q18Surface" (Default=Q7Steane)',
    )
    parser.add_argument(
        "-fq",
        type=int,
        default=100,
        help="Specify after how many qubit usages error correction is applied to it (Default=100)",
    )

    args = parser.parse_args()

    error_channels = args.m
    error_probability = args.p
    number_of_shots = args.n
    seed = args.s
    open_qasm_file = args.f

    if args.fs.lower() == "none":
        forced_simulator = None
    else:
        forced_simulator = args.fs

    if args.ecc.lower() == "none":
        ecc = None
    else:
        ecc = args.ecc

    ecc_frequency = args.fq
    ecc_export_filename = args.e

    if "stabilizer" in forced_simulator and "A" in error_channels:
        print(
            'Warning: Non-unitary errors (such as for example amplitude damping ("A")) are not suitable for simulation '
            "with a stabilizer based simulator and may cause an error during the simulation."
        )

    # Creating the noise model
    if error_probability > 0:
        noise_model = create_noise_model(n_model=error_channels, p_error=error_probability)
    else:
        noise_model = NoiseModel()

    circ = QuantumCircuit.from_qasm_file(open_qasm_file)

    if not any(gate[0].name == "measure" for gate in circ.data):
        print("Warning: No measurement gates found in the circuit. Adding measurement gates to all qubits.")
        circ.measure_all()

    # Initializing the quantum circuit
    if ecc is not None:
        # Applying error correction to the circuit
        result = qecc.apply_ecc(circ, ecc, ecc_frequency)
        circ = QuantumCircuit().from_qasm_str(result["circ"])

    if ecc_export_filename is not None:
        print("Exporting circuit to: " + str(ecc_export_filename))
        circ.qasm(filename=ecc_export_filename)
        return

    size = circ.num_qubits
    simulator_backend = None
    print(
        "_____Trying to simulate with "
        + str(error_channels)
        + " (prob="
        + str(error_probability)
        + ", shots="
        + str(number_of_shots)
        + ", n_qubits="
        + str(size)
        + ", error correction="
        + str(ecc)
        + ") Error______",
        flush=True,
    )

    # Setting the simulator backend to the requested one
    try:
        simulator_backend = Aer.get_backend(forced_simulator)
    except providers.exceptions.QiskitBackendNotFoundError:
        raise ValueError(
            "Simulator " + str(forced_simulator) + " not found! Available simulators are: " + str(Aer.backends())
        ) from None

    result = execute(
        circ,
        backend=simulator_backend,
        shots=number_of_shots,
        seed_simulator=seed,
        noise_model=noise_model,
    )

    if result.result().status != "COMPLETED":
        raise RuntimeError("Simulation exited with status: " + str(result.result().status))

    result_counts = result.result().get_counts()
    print_simulation_results(result_counts, number_of_shots)



import json
from pathlib import Path

import matplotlib
import matplotlib.colors as mcolors
import matplotlib.pyplot as plt
import numpy as np
import scipy.optimize as opt

matplotlib.rcParams["mathtext.fontset"] = "custom"
matplotlib.rcParams["mathtext.rm"] = "Bitstream Vera Sans"
matplotlib.rcParams["mathtext.it"] = "Bitstream Vera Sans:italic"
matplotlib.rcParams["mathtext.bf"] = "Bitstream Vera Sans:bold"


def lin_fun(x, a, b):
    return np.multiply(a, x) + b


def quad_fun(x, a, b):
    return np.multiply(a, np.power(x, 2)) + b


def pow_three_fun(x, a, b):
    return np.multiply(a, np.power(x, 3)) + b


def pow_four_fun(x, a, b):
    return np.multiply(a, np.power(x, 4)) + b


def runtime():
    plt.rcParams.update({"font.size": 15})
    input_filen = "/home/luca/Documents/codeRepos/qecc/scripts/numerics/data/runtime/rt-original-1k-01.json"
    x_data = []
    y_data = []
    pers = []
    orders = []

    with Path(input_filen).open as data_file:
        data = json.load(data_file)

    for per in data:
        per_x_data = []
        per_y_data = []

        for c in data[per]:
            per_x_data.append(float(c))
            per_y_data.append(float(data[per][c]) / 1000.0)
        pers.append(float(per))
        x_data.append(per_x_data)
        y_data.append(per_y_data)

    xfinal = []
    yfinal = []
    cols = []
    colors = mcolors.BASE_COLORS

    for i in range(len(x_data)):
        col, val = colors.popitem()
        if col == "w" or col == "k":
            col, val = colors.popitem()
            if col == "w" or col == "k":
                col, val = colors.popitem()
                cols.append(col)
        label = "% 6.3f" % pers[i]
        orders.append(np.argsort(x_data[i]))
        xfinal.append(np.array(x_data[i])[orders[i]])
        yfinal.append(np.array(y_data[i])[orders[i]])
        plt.plot(xfinal[i], yfinal[i], "o", label="p=" + label, color=col)
        # start = 5
        # optimizedParameters, pcov = opt.curve_fit(LinFun, xfinal[i][start:], yfinal[i][start:])
        # plt.plot(xfinal[i][start:], LinFun(xfinal[i][start:], *optimizedParameters), '--', color=col, label='O(n)')
        print(xfinal)
        print(yfinal)

    # start2 = 7
    # optimizedParameters2, pcov = opt.curve_fit(LinFun, xfinal[1][start2:], yfinal[1][start2:])
    # plt.plot(xfinal[1][start2:], LinFun(xfinal[1][start2:], *optimizedParameters2),'--', color='m', label='O(n)')

    plt.legend()
    plt.ylabel("avg runtime (s) to decode 1k samples")
    plt.xlabel("code length n")
    plt.grid()
    plt.show()


def runtime_comparison():
    plt.rcParams.update({"font.size": 14})
    input_filen = ""
    input_filen2 = ""
    x_data = []
    y_data = []
    pers = []
    x_data2 = []
    y_data2 = []
    pers2 = 0.0
    orders = []

    with Path(input_filen).open() as data_file:
        data = json.load(data_file)
    with Path(input_filen2).open() as data_file2:
        data2 = json.load(data_file2)

    for per in data:
        per_x_data = []
        per_y_data = []

        for c in data[per]:
            per_x_data.append(float(c))
            per_y_data.append(float(data[per][c]) / 1000.0)
        pers.append(float(per))
        x_data.append(per_x_data)
        y_data.append(per_y_data)

    for per in data2:

        for c in data2[per]:
            x_data2.append(float(c))
            y_data2.append(float(data2[per][c]) / 1000.0)
        pers2 = float(per)

    xfinal = []
    yfinal = []
    cols = []
    colors = mcolors.BASE_COLORS

    for i in range(len(x_data)):
        col, val = colors.popitem()
        if col == "w" or col == "k":
            col, val = colors.popitem()
            if col == "w" or col == "k":
                col, val = colors.popitem()
                cols.append(col)
        label = "%2.2f" % pers[i]
        orders.append(np.argsort(x_data[i]))
        xfinal.append(np.array(x_data[i])[orders[i]])
        yfinal.append(np.array(y_data[i])[orders[i]])
        plt.plot(xfinal[i], yfinal[i], "o", label="UFH, p=" + label, color=col)
        start = 0
        optimized_parameters, pcov = opt.curve_fit(lin_fun, xfinal[i][start:], yfinal[i][start:])
        plt.plot(xfinal[i][start:], lin_fun(xfinal[i][start:], *optimized_parameters), "--", color=col, label=r"$O(n)$")
        print(xfinal)
        print(yfinal)

    # general qlpd decoder data
    label = "%2.2f" % pers2
    orders2 = np.argsort(x_data2)
    xfinal2 = np.array(x_data2)[orders2]
    yfinal2 = np.array(y_data2)[orders2]
    plt.plot(xfinal2, yfinal2, "d", label="GD, p=" + label, color="green")
    start = 0
    optimized_parameters2, pcov = opt.curve_fit(quad_fun, xfinal2[start:], yfinal2[start:])
    plt.plot(
        xfinal2[start:], quad_fun(xfinal2[start:], *optimized_parameters2), "--", color="green", label=r"$O(n^{2})$"
    )

    plt.legend()
    plt.ylabel("avg runtime (s) to decode $10^3$ samples")
    plt.xlabel("code length n")
    plt.grid()
    plt.show()


runtime_comparison()



# taken from https://github.com/quantumgizmos/bias_tailored_qldpc
import ldpc.protograph as pt
import numpy as np
from bposd.css import css_code
from bposd.stab import stab_code


def Identity(n):
    return pt.identity(n)


class lifted_hgp(css_code):
    def __init__(self, lift_parameter, a, b=None):

        """
        Generates the lifted hypergraph product of the protographs a and b
        """
        self.a = a

        self.a_m, self.a_n = self.a.shape

        if b is None:
            self.b = pt.copy(self.a)
        else:
            self.b = b

        self.b_m, self.b_n = self.b.shape

        self.hx1_proto = np.kron(self.a, Identity(self.b_n))
        self.hx2_proto = np.kron(Identity(self.a_m), self.b.T)
        self.hx_proto = pt.hstack([self.hx1_proto, self.hx2_proto])

        self.hz1_proto = np.kron(Identity(self.a_n), self.b)
        self.hz2_proto = np.kron(self.a.T, Identity(self.b_m))
        self.hz_proto = pt.hstack([self.hz1_proto, self.hz2_proto])

        self.lift_parameter = lift_parameter

        super().__init__(self.hx_proto.to_binary(lift_parameter), self.hz_proto.to_binary(lift_parameter))

    @property
    def protograph(self):
        px = pt.vstack([pt.zeros(self.hz_proto.shape), self.hx_proto])
        pz = pt.vstack([self.hz_proto, pt.zeros(self.hx_proto.shape)])
        return pt.hstack([px, pz])

    @property
    def hx1(self):
        return self.hx1_proto.to_binary(self.lift_parameter)

    @property
    def hx2(self):
        return self.hx2_proto.to_binary(self.lift_parameter)

    @property
    def hz1(self):
        return self.hz1_proto.to_binary(self.lift_parameter)

    @property
    def hz2(self):
        return self.hz2_proto.to_binary(self.lift_parameter)


class bias_tailored_lifted_product(stab_code):
    def __init__(self, lift_parameter, a, b=None):
        lhgp = lifted_hgp(lift_parameter, a, b)

        # Hadamard rotation
        temp1 = pt.hstack([pt.zeros(lhgp.hx1_proto.shape), lhgp.hz2_proto])
        temp2 = pt.hstack([lhgp.hx1_proto, pt.zeros(lhgp.hz2_proto.shape)])
        self.hx_proto = pt.vstack([temp1, temp2])
        temp1 = pt.hstack([lhgp.hz1_proto, pt.zeros(lhgp.hx2_proto.shape)])
        temp2 = pt.hstack([pt.zeros(lhgp.hz1_proto.shape), lhgp.hx2_proto])
        self.hz_proto = pt.vstack([temp1, temp2])

        super().__init__(self.hx_proto.to_binary(lift_parameter), self.hz_proto.to_binary(lift_parameter))

    @property
    def protograph(self):
        px = pt.vstack([pt.zeros(self.hz_proto.shape), self.hx_proto])
        pz = pt.vstack([self.hz_proto, pt.zeros(self.hx_proto.shape)])
        return pt.hstack([px, pz])



import json
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np


def wer():
    plt.rcParams.update({"font.size": 14})
    input_filename = ""
    input_filename2 = ""
    input_filename3 = ""
    fig, ax = plt.subplots(1)

    with Path(input_filename).open() as f:
        data = json.load(f)
    x_data = []
    y_data = []
    for key in data:
        x_data.append(float(key))
        y_data.append(data[key])

    order = np.argsort(x_data)
    x_dataf = np.array(x_data)[order]
    y_dataf = np.array(y_data)[order]

    with Path(input_filename2).open() as f:
        data2 = json.load(f)
    x_data2 = []
    y_data2 = []
    for key2 in data2:
        x_data2.append(float(key2))
        y_data2.append(data2[key2])

    order2 = np.argsort(x_data2)
    x_data2f = np.array(x_data2)[order2]
    y_data2f = np.array(y_data2)[order2]

    with Path(input_filename3).open() as f:
        data3 = json.load(f)
    x_data3 = []
    y_data3 = []
    for key3 in data3:
        x_data3.append(float(key3))
        y_data3.append(data3[key3])

    order3 = np.argsort(x_data3)
    x_data3f = np.array(x_data3)[order3]
    y_data3f = np.array(y_data3)[order3]

    ax.plot(x_dataf, y_dataf, "-x", label="heuristic SSG", color="b")
    ax.plot(x_data2f, y_data2f, "-o", label="GD", color="g")
    ax.plot(x_data3f, y_data3f, "-x", label="heuristic AG", color="r")
    print(y_data)
    print(y_data2)
    print(y_data3)

    ax.set_xlabel("physical X-error rate")
    ax.set_ylabel("WER")
    ax.legend()
    ax.grid()
    ax.set_xscale("log")
    ax.set_yscale("log")
    # ax.gca().set_position([0, 0, 1, 1])
    handles, labels = ax.get_legend_handles_labels()

    handles = [handles[1], handles[2], handles[0]]
    labels = [labels[1], labels[2], labels[0]]

    ax.legend(handles, labels, loc=4)
    plt.show()


def wer_comp():
    plt.rcParams.update({"font.size": 15})
    input_filename = ""
    input_filename2 = ""
    fig, ax = plt.subplots(1)

    with Path(input_filename).open() as f:
        data = json.load(f)
    x_data = []
    y_data = []
    for key in data:
        x_data.append(float(key))
        y_data.append(data[key])

    with Path(input_filename2).open() as f:
        data2 = json.load(f)
    x_data2 = []
    y_data2 = []
    for key2 in data2:
        x_data2.append(float(key2))
        y_data2.append(data2[key2])

    ax.plot(x_data, y_data, "-d", label="heuristic", color="b")
    ax.plot(x_data2, y_data2, "-o", label="decoder", color="g")
    print(y_data)
    print(y_data2)

    ax.xlabel("physical X-error rate")
    ax.ylabel("WER")
    ax.legend()
    ax.grid()
    ax.xscale("log")
    ax.yscale("log")
    ax.gca().set_position([0, 0, 1, 1])
    ax.savefig("werComp.svg")
    handles, labels = ax.get_legend_handles_labels()

    handles = [handles[2], handles[1], handles[0]]
    labels = [labels[2], labels[1], labels[0]]

    ax.legend(handles, labels, loc=2)
    plt.show()



import os
import re
import subprocess
import sys
from contextlib import suppress
from pathlib import Path

from setuptools import Extension, setup
from setuptools.command.build_ext import build_ext


class CMakeExtension(Extension):
    def __init__(self, name, sourcedir=""):
        Extension.__init__(self, name, sources=[])
        self.sourcedir = str(Path(sourcedir).resolve())


class CMakeBuild(build_ext):
    def build_extension(self, ext):
        from setuptools_scm import get_version

        version = get_version(relative_to=__file__)

        extdir = str(Path(self.get_ext_fullpath(ext.name)).parent.resolve())

        cmake_generator = os.environ.get("CMAKE_GENERATOR", "")
        cfg = "Debug" if self.debug else "Release"

        cmake_args = [
            f"-DCMAKE_LIBRARY_OUTPUT_DIRECTORY={extdir}",
            f"-DPYTHON_EXECUTABLE={sys.executable}",
            f"-DQECC_VERSION_INFO={version}",
            f"-DCMAKE_BUILD_TYPE={cfg}",
            "-DBINDINGS=ON",
        ]
        build_args = []

        if self.compiler.compiler_type != "msvc":
            if not cmake_generator:
                cmake_args += ["-GNinja"]
        else:
            # Single config generators are handled "normally"
            single_config = any(x in cmake_generator for x in {"NMake", "Ninja"})
            # CMake allows an arch-in-generator style for backward compatibility
            contains_arch = any(x in cmake_generator for x in {"ARM", "Win64"})
            # Convert distutils Windows platform specifiers to CMake -A arguments
            plat_to_cmake = {
                "win32": "Win32",
                "win-amd64": "x64",
                "win-arm32": "ARM",
                "win-arm64": "ARM64",
            }
            # Specify the arch if using MSVC generator, but only if it doesn't
            # contain a backward-compatibility arch spec already in the
            # generator name.
            if not single_config and not contains_arch:
                cmake_args += ["-A", plat_to_cmake[self.plat_name]]
            # Multi-config generators have a different way to specify configs
            if not single_config:
                cmake_args += [f"-DCMAKE_LIBRARY_OUTPUT_DIRECTORY_{cfg.upper()}={extdir}"]
                build_args += ["--config", cfg]

        # cross-compile support for macOS - respect ARCHFLAGS if set
        if sys.platform.startswith("darwin"):
            archs = re.findall(r"-arch (\S+)", os.environ.get("ARCHFLAGS", ""))
            if archs:
                cmake_args += ["-DCMAKE_OSX_ARCHITECTURES={}".format(";".join(archs))]

        # Set CMAKE_BUILD_PARALLEL_LEVEL to control the parallel build level across all generators.
        if "CMAKE_BUILD_PARALLEL_LEVEL" not in os.environ and hasattr(self, "parallel") and self.parallel:
            build_args += [f"-j{self.parallel}"]

        if sys.platform == "win32":
            cmake_args += ["-T", "ClangCl"]

        build_dir = Path(self.build_temp)
        build_dir.mkdir(parents=True, exist_ok=True)
        with suppress(FileNotFoundError):
            Path(build_dir / "CMakeCache.txt").unlink()

        subprocess.check_call(["cmake", ext.sourcedir] + cmake_args, cwd=self.build_temp)
        subprocess.check_call(
            ["cmake", "--build", ".", "--target", ext.name.split(".")[-1]] + build_args,
            cwd=self.build_temp,
        )


setup(
    ext_modules=[CMakeExtension("mqt.qecc.pyqecc")],
    cmdclass={"build_ext": CMakeBuild},
)



from __future__ import annotations

import os

import nox
from nox.sessions import Session

nox.options.sessions = ["lint", "tests"]

PYTHON_ALL_VERSIONS = ["3.7", "3.8", "3.9", "3.10"]

if os.environ.get("CI", None):
    nox.options.error_on_missing_interpreters = True


@nox.session(python=PYTHON_ALL_VERSIONS)
def tests(session: Session) -> None:
    """
    Run the test suite.
    Simply execute `nox -rs tests` to run all tests.
    Run as `nox -rs tests -- skip-install` to skip installing the package and its dependencies.
    """
    run_install = True
    if session.posargs and "skip-install" in session.posargs:
        run_install = False
        session.posargs.remove("skip-install")
    if run_install:
        session.install("numpy")
        session.install("-e", ".[test]")
    session.run("pytest", *session.posargs)


@nox.session(python=PYTHON_ALL_VERSIONS)
def coverage(session: Session) -> None:
    """
    Run the test suite and generate a coverage report.
    Simply execute `nox -rs coverage -- --cov-report=html` to generate a HTML report.
    Run as `nox -rs coverage -- skip-install` to skip installing the package and its dependencies.
    """
    run_install = True
    if session.posargs and "skip-install" in session.posargs:
        run_install = False
        session.posargs.remove("skip-install")
    if run_install:
        session.install("numpy")
        session.install("-e", ".[coverage]")
    session.run("pytest", "--cov", *session.posargs)


@nox.session
def lint(session: Session) -> None:
    """
    Lint the Python part of the codebase using pre-commit.
    Simply execute `nox -rs lint` to run all configured hooks.
    """
    session.install("pre-commit")
    session.run("pre-commit", "run", "--all-files", *session.posargs)


@nox.session
def pylint(session: Session) -> None:
    """
    Run pylint.
    Simply execute `nox -rs pylint` to run pylint.
    Run as `nox -rs pylint -- skip-install` to skip installing the package and its dependencies.
    """
    session.install("pylint")
    run_install = True
    if session.posargs and "skip-install" in session.posargs:
        run_install = False
        session.posargs.remove("skip-install")
    if run_install:
        session.install("numpy")
        session.install("-e", ".")
    session.run("pylint", "mqt.qecc", "--extension-pkg-allow-list=mqt.qecc.pyqecc", *session.posargs)


@nox.session
def mypy(session: Session) -> None:
    """
    Run mypy.
    Simply execute `nox -rs mypy` to run mypy.
    """
    session.install("pre-commit")
    session.run("pre-commit", "run", "--all-files", "--hook-stage", "manual", "mypy", *session.posargs)


@nox.session
def docs(session: Session) -> None:
    """
    Build the documentation.
    Simply execute `nox -rs docs -- serve` to locally build and serve the docs.
    Run as `nox -rs docs -- skip-install` to skip installing the package and its dependencies.
    """
    run_install = True
    if session.posargs and "skip-install" in session.posargs:
        run_install = False
        session.posargs.remove("skip-install")
    if run_install:
        session.install("-e", ".[docs]")
    session.chdir("docs")
    session.run("sphinx-build", "-M", "html", "source", "_build")

    if session.posargs:
        if "serve" in session.posargs:
            print("Launching docs at http://localhost:8000/ - use Ctrl-C to quit")
            session.run("python", "-m", "http.server", "8000", "-d", "_build/html")
        else:
            print("Unsupported argument to docs")



from pathlib import Path

from mqt import qecc

codes = [
    "../../examples/toricCodes/toric_(nan,nan)-[[8,2,2]]_hz.txt",
    "../../examples/toricCodes/toric_(nan,nan)-[[18,2,3]]_hz.txt",
    "../../examples/toricCodes/toric_(nan,nan)-[[32,2,4]]_hz.txt",
    "../../examples/toricCodes/toric_(nan,nan)-[[50,2,5]]_hz.txt",
    "../../examples/toricCodes/toric_(nan,nan)-[[72,2,6]]_hz.txt",
    "../../examples/toricCodes/toric_(nan,nan)-[[98,2,7]]_hz.txt",
    "../../examples/toricCodes/toric_(nan,nan)-[[128,2,8]]_hz.txt",
    "../../examples/toricCodes/toric_(nan,nan)-[[162,2,9]]_hz.txt",
    "../../examples/toricCodes/toric_(nan,nan)-[[200,2,10]]_hz.txt",
]
outpath = "./rt-sims-bindings.out"

nrSamples = 1
nrRuns = 1
per = 0.01

with Path(outpath).open("w") as outfile:
    for codePath in codes:
        sampleSum = 0.0
        for _ in range(nrSamples):
            runsSum = 0.0
            for _ in range(nrRuns):
                code = qecc.Code(codePath)
                err = qecc.sample_iid_pauli_err(code.N, per)
                decoder = qecc.UFHeuristic()
                decoder.set_code(code)
                syndr = code.get_x_syndrome(err)
                decoder.decode(syndr)
                time = decoder.result.decoding_time
                runsSum += time
            sampleSum += runsSum
        outp = codePath + ":" + str(sampleSum / nrSamples)
        print(outp)
        outfile.write(outp + "\n")
        outfile.flush()



from pathlib import Path

from mqt import qecc

code = qecc.Code("../../examples/lp_(4,8)-[[1024,18,nan]]_hx.txt", "../../examples/lp_(4,8)-[[1024,18,nan]]_hz.txt")
code.K = 18
outpath = "./dp-sims-bindings.out"

runsPerEr = 1
currPer = 0.00001
maxPer = 0.00003
stepSize = 0.00001
nrFailedRuns = 0
codeK = code.K

with Path(outpath).open("w") as outfile:
    while currPer < maxPer:
        nrFailedRuns = 0
        for _ in range(runsPerEr):
            decoder = qecc.UFHeuristic()
            decoder.set_code(code)
            err = qecc.sample_iid_pauli_err(code.N, currPer)
            decoder.decode(code.get_x_syndrome(err))
            result = decoder.result.estimate
            if not code.is_stabilizer(result):
                nrFailedRuns += 1
            currPer += stepSize
        fracFailed = nrFailedRuns / runsPerEr
        wer = fracFailed / codeK
        res = str(currPer) + ":" + str(wer) + "\n"
        print(res)
        outfile.write(res)
        currPer += stepSize
